/*
  Copyright (c) 2008 Seneca College
  Licenced under the MIT License (http://www.c3dl.org/index.php/mit-license/)
  Author: Patrick Lam
*/

/**
	@class A Picking class
*/
function Picking(scene)
{
	var debug = "";
	var debugField;
	var scn = new Scene();
	scn = scene.getScene();
	var cam = new FreeCamera();
	cam = scn.getCamera();

	this.setScene = function(scene)
	{
		scn = scene;
	}
	
	// ====================================================
	// For debugging
	function displayDebug(message) {
		debugField.value = message;
	}

	this.setTextField = function(field) {
		debugField = field;
	}
	// ====================================================

	this.onMouseUp = function(event)
	{
	}

	/**
		Returns an array of object index number in the current scene in sorted order (closet to furthest)

		@returns {Array} The array of object index number.
	*/
	this.onMouseDown = function(event)
	{
		debug = "";
		// Button    Which    Button
		// Left  	 1		  0
		// Middle 	 2		  1
		// Right 	 3  	  2

		//==========================================================================
		// Disables Right Click Menu - copied from internet
		var message="Right Clicked";

		function clickIE(){
			if (document.all){
				(message);
				return false;
			}
		}

		function clickNS(e) {
			if(document.layers||(document.getElementById&&!document.all)){
				if (e.which==2||e.which==3) {
					alert(message);
					return false;
				}
			}
		}

		if (document.layers){
			document.captureEvents(Event.MOUSEDOWN);
			document.onmousedown = clickNS;
		}
		else{
			document.onmouseup = clickNS;
			document.oncontextmenu = clickIE;
		}

		document.oncontextmenu = new Function("return false")
		//==========================================================================

		// getElementsByTagName returns an array
		// Take the first canvas tag
		var canvasTag = document.getElementsByTagName('canvas')[0];

		// Mouse Coord on canvas
		var canvasXY = getXandY(event);

		// Camera Position
		var camPos = cam.getPosition();

		// Find dimension of far clipping plane
		var farWidth = 2 * ((C3DL_FAR_CLIPPING_PLANE) * Math.tan(degreesToRadians(0.5 * C3DL_FIELD_OF_VIEW)));
		var farHeight = farWidth * (canvasTag.height/canvasTag.width);

		// Find ratio between far clipping plane and canvas screen
		var wRatio = farWidth / canvasTag.width;
		var hRatio = farHeight / canvasTag.width;

		// Find point clicked on far clipping plane
		var farX = (canvasXY[0] / canvasTag.width) * farWidth;
		var farY = (canvasXY[1] / canvasTag.height) * farHeight;
		var camDist = Math.sqrt(camPos[0] * camPos[0] + camPos[1] * camPos[1] + camPos[2] * camPos[2]);
		// Since we are starting off with positions aligned on the z-axis
		// The camera is hypothetically on the z-axis, and so, the z value should be
		// the distance of the camera from the origin of the axis and not just the
		// the z value (projection on z-axis) of the camera's position
		var farZ = C3DL_FAR_CLIPPING_PLANE - camDist;

		// Create the Starting point and direction of mouse vector
		var mouseOrigin = camPos;
		var mouseVec = makeVector(farX, farY, farZ);

		// View Matrix
		viewMatrix = makePoseMatrix(cam.getLeft(), cam.getUp(), cam.getDir(), cam.getPosition());

		// Move mouse vector to align with camera
		mouseVec = multiplyMatrixByVector(viewMatrix, mouseVec);

		// Array holding index of objects that pass the bounding box test
		var passRayBoxTest = new Array();

		// Find all objects with bounding box intersecting with mouse vector
		for (var i = 0; i < scn.getObjListSize(); i++) {
			var testObj = scn.getObj(i);
			var boxValues = testObj.getBoundingBox(true);
			// getBoundingBox() returns {minX, minY, minZ, maxX, maxY, maxZ} that has not been transformed
			// only gets scaled and object's position offset
			var minVal = new Array(boxValues[0], boxValues[1], boxValues[2]);
			var maxVal = new Array(boxValues[3], boxValues[4], boxValues[5]);

			// Move the mouse vector into the object's space
			var objTransMat = makePoseMatrix(testObj.getLeft(), testObj.getUp(), testObj.getDirection(), testObj.getPosition());
			var objInvTransMat = inverseMatrix(objTransMat);
			var tempOrigin = multiplyMatrixByVector(objInvTransMat, mouseOrigin);
			var tempVec = multiplyMatrixByVector(objInvTransMat, mouseVec);

			if (CheckLineBox(minVal, maxVal, tempOrigin, tempVec)) {
				passRayBoxTest.push(i);
			}
		}

		// ===========================================================================
		//	FOR RAY-TRIANGLE INTERSECTION TEST, STILL WORKING ON IT
		// ===========================================================================
		// Array holding index of objects that pass the triangle test
		var intersectObjs = new Array();

		// Find all objects with triangles(faces) intersecting with mouse vector
		for (var i = 0; i < passRayBoxTest.length; i++) {
			var test2Obj = scn.getObj(passRayBoxTest[i]);
			var vertices = test2Obj.getVertices();
			var passTest = false;	// flag for when to stop triangle test once passed
			
			// Object's scale
			var wtScale = new Array( test2Obj.getScale()[0],0,0,0,
									 0,test2Obj.getScale()[1],0,0,
									 0,0,test2Obj.getScale()[2],0,
									 0,0,0,1);	

			// Move the mouse vector into the object's space
			// This avoids applying the objects transformation matrix to all it's vertices for the test
			var objTransMat2 = makePoseMatrix(test2Obj.getLeft(), test2Obj.getUp(), test2Obj.getDirection(), test2Obj.getPosition());
			var objInvTransMat2 = inverseMatrix(objTransMat2);
			var tempOrigin2 = multiplyMatrixByVector(objInvTransMat2, mouseOrigin);
			var tempVec2 = multiplyMatrixByVector(objInvTransMat2, mouseVec);

			// Go through the object's faces
			for (var j = 0; j < vertices.length && !passTest; j += 9) {
				// 3 points of a triangle with the object's position offset
				var vert1 = new Array(vertices[j], vertices[j+1], vertices[j+2]);
				var vert2 = new Array(vertices[j+3], vertices[j+4], vertices[j+5]);
				var vert3 = new Array(vertices[j+6], vertices[j+7], vertices[j+8]);

				// Scale the vertices
				vert1 = multiplyMatrixByVector(wtScale, vert1);
				vert2 = multiplyMatrixByVector(wtScale, vert2);
				vert3 = multiplyMatrixByVector(wtScale, vert3);

				// Add on object's position offset
				vert1 = addVectors(vert1,  test2Obj.getPosition());
				vert2 = addVectors(vert2,  test2Obj.getPosition());
				vert3 = addVectors(vert3,  test2Obj.getPosition());

				var result = intersect_triangle(tempOrigin2, tempVec2, vert1, vert2, vert3);
				// Adding objects that pass triangle test
				if (result == true) {
					intersectObjs.push(passRayBoxTest[i]);
					passTest = true;
				}
			}
		}

		// Sort all intersecting objects from closet to farthest 
		for (var i = 0; i < intersectObjs.length; i++) {
			for (var j = 0; j < intersectObjs.length; j++) {
				// Objects Positions
				var objAPos = scn.getObj(intersectObjs[i]).getPosition();
				var objBPos = scn.getObj(intersectObjs[j]).getPosition();

				// Objects x,y,z distances
				var dxA = camPos[0] - objAPos[0];
				var dyA = camPos[1] - objAPos[1];
				var dzA = camPos[2] - objAPos[2];

				var dxB = camPos[0] - objBPos[0];
				var dyB = camPos[1] - objBPos[1];
				var dzB = camPos[2] - objBPos[2];

				// Objects distance from camera
				var distA = Math.sqrt(dxA * dxA + dyA * dyA + dzA * dzA);
				var distB = Math.sqrt(dxB * dxB + dyB * dyB + dzB * dzB);

				// Swap
				if (distA < distB) {
					var tmp = intersectObjs[i];
					intersectObjs[i] = intersectObjs[j];
					intersectObjs[j] = tmp;	
				}
			}
		}
		return intersectObjs;
	}

	// -------------------------------------------------------------------------
	// Mouse Position Functions

	// Find position of object in browser
	// Returning it's offset in the browser
	function findPos(obj) {
		var curleft = curtop = 0;
		if (obj.offsetParent) {
			do {
				curleft += obj.offsetLeft;
				curtop += obj.offsetTop;
			} while (obj = obj.offsetParent);
			return [curleft,curtop];
		}
	}

	// Returns the mouse x,y in canvas
	function getXandY( event )
	{
		var tempDiv = document.getElementsByTagName('canvas')[0];
		var coords = findPos(tempDiv);

		// determine the correct X, Y
		var X = ( ( event.clientX - coords[0] ) + window.pageXOffset )  - (tempDiv.width / 2.0);
		var Y = (tempDiv.height / 2.0) - ( event.clientY - ( coords[1] - window.pageYOffset ) );

		return [X,Y];
	}

	// -------------------------------------------------------------------------
	// Ray-Bounding Box Intersection Test
	// Modified Version of Sample function from 3D Programming
	function GetIntersection(fDst1, fDst2, P1, P2) {
		if ((fDst1 * fDst2) >= 0) return 0;
		if ( fDst1 == fDst2 ) return 0;

		var Hit = new Array();

		Hit[0] = P1[0] + (P2[0]-P1[0]) * ( (-1.0 * fDst1) / (fDst2 - fDst1) );
		Hit[1] = P1[1] + (P2[1]-P1[1]) * ( (-1.0 * fDst1) / (fDst2 - fDst1) );
		Hit[2] = P1[2] + (P2[2]-P1[2]) * ( (-1.0 * fDst1) / (fDst2 - fDst1) );

		return Hit;
	}

	function InBox( Hit, B1, B2, Axis) {
		if (Hit != 0) {
			if ( Axis==1 && Hit[2] > B1[2] && Hit[2] < B2[2] && Hit[1] > B1[1] && Hit[1] < B2[1]) return 1;
			if ( Axis==2 && Hit[2] > B1[2] && Hit[2] < B2[2] && Hit[0] > B1[0] && Hit[0] < B2[0]) return 1;
			if ( Axis==3 && Hit[0] > B1[0] && Hit[0] < B2[0] && Hit[1] > B1[1] && Hit[1] < B2[1]) return 1;
		}	

		return 0;
	}

	// B1 is array of 3 values, the min X,Y,Z values of bounding box
	// B2 is array of 3 values, the max X,Y,Z values of bounding box
	// L1 is array of 3 values, the X,Y,Z values of starting position of ray
	// L2 is array of 3 values, the X,Y,Z values of direction vector of ray	
	function CheckLineBox(B1, B2, L1, L2) {
		if (L2[0] < B1[0] && L1[0] < B1[0]) return false;
		if (L2[0] > B2[0] && L1[0] > B2[0]) return false;
		if (L2[1] < B1[1] && L1[1] < B1[1]) return false;
		if (L2[1] > B2[1] && L1[1] > B2[1]) return false;
		if (L2[2] < B1[2] && L1[2] < B1[2]) return false;
		if (L2[2] > B2[2] && L1[2] > B2[2]) return false;
		if (L1[0] > B1[0] && L1[0] < B2[0] &&
			L1[1] > B1[1] && L1[1] < B2[1] &&
			L1[2] > B1[2] && L1[2] < B2[2]) 
			{return true;}
		if ( (InBox( GetIntersection( L1[0]-B1[0], L2[0]-B1[0], L1, L2), B1, B2, 1 ))
		  || (InBox( GetIntersection( L1[1]-B1[1], L2[1]-B1[1], L1, L2), B1, B2, 2 )) 
		  || (InBox( GetIntersection( L1[2]-B1[2], L2[2]-B1[2], L1, L2), B1, B2, 3 )) 
		  || (InBox( GetIntersection( L1[0]-B2[0], L2[0]-B2[0], L1, L2), B1, B2, 1 )) 
		  || (InBox( GetIntersection( L1[1]-B2[1], L2[1]-B2[1], L1, L2), B1, B2, 2 )) 
		  || (InBox( GetIntersection( L1[2]-B2[2], L2[2]-B2[2], L1, L2), B1, B2, 3 )))
		  {
			return true;
		  }

		return false;
	}
	// -------------------------------------------------------------------------
	// Ray-Triangle Intersection Test
	function intersect_triangle(orig, dir, vert0, vert1, vert2) {
		/* find vectors for two edges sharing vert0 */
		var edge1 = new Array(vert1[0] - vert0[0], vert1[1] - vert0[1], vert1[2] - vert0[2]);
		var edge2 = new Array(vert2[0] - vert0[0], vert2[1] - vert0[1], vert2[2] - vert0[2]);

		// area of triangle
		var area = 0.5 * vectorLength(new Array(edge1[1] * edge2[2] - edge1[2] * edge2[1], edge1[0] * edge2[2] - edge1[2] * edge2[0], edge1[0] * edge2[1] - edge1[1] * edge2[0]));

		// Plane equation is ax + by + cz = d
		// n = [a b c]   x = [x y z]
		// n.x = d
		var norm = new Array(edge1[1] * edge2[2] - edge1[2] * edge2[1], edge1[0] * edge2[2] - edge1[2] * edge2[0], edge1[0] * edge2[1] - edge1[1] * edge2[0]);
		var d = norm[0] * vert1[0] + norm[1] * vert1[1] + norm[2] * vert1[2];

		// Ray is parallel to plane
		var nd = norm[0] * dir[0] + norm[1] * dir[1] + norm[2] * dir[2];
		if (nd == 0) {
			return false;
		}

		// solve for t
		// R(t) = P + (t x dir)
		// t = (d - n.P) / (n.dir)
		var nP = norm[0] * orig[0] + norm[1] * orig[1] + norm[2] * orig[2];
		var t = (d - nP) / nd;
		
		// Point of Intersection of the Ray and the Plane created by the triangle
		var POI = new Array(orig[0] + dir[0] * t, orig[1] + dir[1] * t, orig[2] + dir[2] * t);
		
		// area of smaller triangles formed by the 3 vertices and the point of intersection
		edge1 = new Array(vert0[0] - POI[0], vert0[1] - POI[1], vert0[2] - POI[2]);
		edge2 = new Array(vert1[0] - POI[0], vert1[1] - POI[1], vert1[2] - POI[2]);
		var edge3 = new Array(vert2[0] - POI[0], vert2[1] - POI[1], vert2[2] - POI[2]);

		var area1 = 0.5 * vectorLength(new Array(edge1[1] * edge2[2] - edge1[2] * edge2[1], edge1[0] * edge2[2] - edge1[2] * edge2[0], edge1[0] * edge2[1] - edge1[1] * edge2[0]));
		var area2 = 0.5 * vectorLength(new Array(edge2[1] * edge3[2] - edge2[2] * edge3[1], edge2[0] * edge3[2] - edge2[2] * edge3[0], edge2[0] * edge3[1] - edge2[1] * edge3[0]));
		var area3 = 0.5 * vectorLength(new Array(edge3[1] * edge1[2] - edge3[2] * edge1[1], edge3[0] * edge1[2] - edge3[2] * edge1[0], edge3[0] * edge1[1] - edge3[1] * edge1[0]));

		// Calculate the % of error
		var diff = (area - (area1 + area2 + area3))/area;

		// Check if POI is inside triangle
		// Adjust margin of error for accuracy
		// The threshold that will allow the triangle test work seems to be around 0.6%
		// So the safe range should be around 0.7 - 1.0%
		if (Math.abs(diff) >= 0.8)
			return false;
		else
			return true;
	}
}